""" Test the CLI for run mip-dna """
import logging

from cg.apps.tb import TrailblazerAPI
from cg.cli.workflow.mip_dna.base import run
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI


def test_cg_dry_run(cli_runner, mocker, caplog, case_id, email_adress, dna_mip_context):
    """Test print the MIP run to console"""

    caplog.set_level(logging.INFO)

    mocker.patch.object(MipDNAAnalysisAPI, "get_target_bed_from_lims")
    MipDNAAnalysisAPI.get_target_bed_from_lims.return_value = (
        "tests/fixtures/apps/mip/rna/case_config.yaml"
    )
    mocker.patch.object(TrailblazerAPI, "is_latest_analysis_ongoing")
    TrailblazerAPI.is_latest_analysis_ongoing.return_value = False

    # GIVEN a cli function
    # WHEN we run a case in dry run mode
    result = cli_runner.invoke(
        run, ["--dry-run", "--email", email_adress, case_id], obj=dna_mip_context
    )
    # THEN command is run successfully
    assert result.exit_code == 0



"""Tests for the scout api"""

import logging

import yaml

from cg.apps.scout.scout_export import Variant
from cg.apps.scout.scoutapi import ScoutAPI


def test_get_causative_variants_no_variants(scout_api: ScoutAPI, case_id: str):
    """Test to get causative variants when there are no variants"""
    # GIVEN a scout api and a process that returns no std output
    assert scout_api.process.stdout == ""

    # WHEN querying for causative variants
    result = scout_api.get_causative_variants(case_id=case_id)

    # THEN assert that the empty list is returned since there where no variants
    assert result == []


def test_get_causative_variants_one_variant(
    scout_api: ScoutAPI, case_id: str, causative_output: str
):
    """Test to get causative variants when there is one variant"""
    scout_api.process.set_stdout(causative_output)
    # GIVEN a scout api and a process that returns some relevant input
    assert scout_api.process.stdout == causative_output

    # WHEN querying for causative variants
    result = scout_api.get_causative_variants(case_id=case_id)

    # THEN assert that the output is a empty list
    assert isinstance(result, list)

    # THEN assert that there was one variant in the list
    assert len(result) == 1
    first_variant: Variant = result[0]
    raw_info = yaml.safe_load(causative_output)

    # THEN assert that the variant has a variant_id
    assert first_variant.document_id == raw_info[0]["_id"]


def test_get_causative_variants_non_existing_case(
    scout_api: ScoutAPI, case_id: str, causative_output: str, caplog
):
    """Test to get causative variants when the case does not exist"""
    caplog.set_level(logging.INFO)
    # GIVEN a scout api and a process that will fail
    scout_api.process.set_exit_code(1)

    # WHEN querying for causative variants
    result = scout_api.get_causative_variants(case_id=case_id)

    # THEN assert that the output is a empty list
    assert isinstance(result, list)
    # THEN assert that the information is communicated
    assert "Could not find case" in caplog.text



from typing import List, Optional, Any

from pydantic import BaseModel, validator, Field

from cg.constants.gender import Gender


def _get_metric_per_sample_id(sample_id: str, metric_objs: list) -> Any:
    """Get metric for a sample_id from metric object"""
    for metric in metric_objs:
        if sample_id == metric.sample_id:
            return metric.value


class MetricsBase(BaseModel):
    """Definition for elements in deliverables metrics file"""

    header: Optional[str]
    id: str
    input: str
    name: str
    step: str
    value: Any


class DuplicateReads(BaseModel):
    """Definition of duplicate reads metric"""

    sample_id: str
    value: float

    @validator("value", always=True)
    def convert_duplicate_read(cls, value) -> float:
        """Convert raw value from fraction to percent"""
        return value * 100


class GenderCheck(BaseModel):
    """Definition of gender check metric"""

    sample_id: str
    value: str


class MappedReads(BaseModel):
    """Definition of mapped reads metric"""

    sample_id: str
    value: float

    @validator("value", always=True)
    def convert_mapped_read(cls, value) -> float:
        """Convert raw value from fraction to percent"""
        return value * 100


class MeanInsertSize(BaseModel):
    """Definition of insert size metric"""

    sample_id: str
    value: float

    @validator("value", always=True)
    def convert_mean_insert_size(cls, value) -> int:
        """Convert raw value from float to int"""
        return int(value)


class ParsedMetrics(BaseModel):
    """Defines parsed metrics"""

    sample_id: str
    duplicate_reads: float
    mapped_reads: float
    mean_insert_size: int
    predicted_sex: str = Gender.UNKNOWN


class MetricsDeliverables(BaseModel):
    """Specification for a metric general deliverables file"""

    metrics_: List[MetricsBase] = Field(..., alias="metrics")
    sample_ids: Optional[set]
    duplicate_reads: Optional[List[DuplicateReads]]
    mapped_reads: Optional[List[MappedReads]]
    mean_insert_size: Optional[List[MeanInsertSize]]
    predicted_sex: Optional[List[GenderCheck]]
    sample_id_metrics: Optional[List[ParsedMetrics]]

    @validator("sample_ids", always=True)
    def set_sample_ids(cls, _, values: dict) -> set:
        """Set sample_ids gathered from all metrics"""
        sample_ids: List = []
        raw_metrics: List = values.get("metrics_")
        for metric in raw_metrics:
            sample_ids.append(metric.id)
        return set(sample_ids)

    @validator("duplicate_reads", always=True)
    def set_duplicate_reads(cls, _, values: dict) -> List[DuplicateReads]:
        """Set duplicate_reads"""
        duplicate_reads: List = []
        raw_metrics: List = values.get("metrics_")
        for metric in raw_metrics:
            if metric.name == "fraction_duplicates":
                duplicate_reads.append(DuplicateReads(sample_id=metric.id, value=metric.value))
        return duplicate_reads

    @validator("mapped_reads", always=True)
    def set_mapped_reads(cls, _, values: dict) -> List[MappedReads]:
        """Set mapped reads"""
        sample_ids: set = values.get("sample_ids")
        mapped_reads: List = []
        total_sequences: dict = {}
        reads_mapped: dict = {}
        raw_metrics: List = values.get("metrics_")
        for metric in raw_metrics:
            if metric.name == "raw_total_sequences":
                raw_total_sequences = total_sequences.get(metric.id, 0)
                total_sequences[metric.id] = int(metric.value) + raw_total_sequences
            if metric.name == "reads_mapped":
                raw_reads_mapped = reads_mapped.get(metric.id, 0)
                reads_mapped[metric.id] = int(metric.value) + raw_reads_mapped
        for sample_id in sample_ids:
            fraction_mapped_read = reads_mapped[sample_id] / total_sequences[sample_id]
            mapped_reads.append(MappedReads(sample_id=sample_id, value=fraction_mapped_read))
        return mapped_reads

    @validator("mean_insert_size", always=True)
    def set_mean_insert_size(cls, _, values: dict) -> List[MeanInsertSize]:
        """Set mean insert size"""
        mean_insert_size: List = []
        raw_metrics: List = values.get("metrics_")
        for metric in raw_metrics:
            if metric.name == "MEAN_INSERT_SIZE":
                mean_insert_size.append(MeanInsertSize(sample_id=metric.id, value=metric.value))
        return mean_insert_size

    @validator("predicted_sex", always=True)
    def set_predicted_sex(cls, _, values: dict) -> List[GenderCheck]:
        """Set predicted sex"""
        predicted_sex: List = []
        raw_metrics: List = values.get("metrics_")
        for metric in raw_metrics:
            if metric.name == "gender":
                predicted_sex.append(GenderCheck(sample_id=metric.id, value=metric.value))
        return predicted_sex

    @validator("sample_id_metrics", always=True)
    def set_sample_id_metrics(cls, _, values: dict) -> List[ParsedMetrics]:
        """Set parsed sample_id metrics gathered from all metrics"""
        sample_ids: set = values.get("sample_ids")
        sample_id_metrics: list = []
        metric_per_sample_id_map: dict = {
            "duplicate_reads": values.get("duplicate_reads"),
            "mapped_reads": values.get("mapped_reads"),
            "mean_insert_size": values.get("mean_insert_size"),
            "predicted_sex": values.get("predicted_sex"),
        }
        for sample_id in sample_ids:
            metric_per_sample_id: dict = {}
            metric_per_sample_id["sample_id"] = sample_id
            for metric_name, metric_objs in metric_per_sample_id_map.items():
                metric_value: Any = _get_metric_per_sample_id(
                    sample_id=sample_id, metric_objs=metric_objs
                )
                if metric_value:
                    metric_per_sample_id[metric_name] = metric_value
            sample_id_metrics.append(ParsedMetrics(**metric_per_sample_id))
        return sample_id_metrics


def get_sample_id_metric(sample_id_metrics: List[ParsedMetrics], sample_id: str) -> ParsedMetrics:
    """Get parsed metrics for an sample_id"""
    for sample_id_metric in sample_id_metrics:
        if sample_id == sample_id_metric.sample_id:
            return sample_id_metric



from cgmodels.cg.constants import StrEnum


class ObservationAnalysisType(StrEnum):
    WGS: str = "wgs"


class ObservationAnalysisTag(StrEnum):
    PEDIGREE: str = "pedigree"
    CHECK_PROFILE_GBCF: str = "snv-gbcf"
    SNV_VARIANTS: str = "deepvariant"
    SV_VARIANTS: str = "vcf-sv-research"



from cg.models.mip.mip_analysis import MipAnalysis, parse_mip_analysis


def test_instantiate_mip_analysis(mip_analysis_raw: dict):
    """
    Tests raw mip analysis against a pydantic MipAnalysis
    """
    # GIVEN a dictionary with the some metrics

    # WHEN instantiating a MipAnalysis object
    mip_analysis_object = MipAnalysis(**mip_analysis_raw)

    # THEN assert that it was successfully created
    assert isinstance(mip_analysis_object, MipAnalysis)


def test_instantiate_parse_mip_analysis(
    mip_analysis_config_dna_raw: dict, mip_metrics_deliverables_raw: dict, sample_info_dna_raw: dict
):
    """
    Tests parse_mip_analysis
    """
    # GIVEN a dictionary with the some metrics

    # WHEN instantiating a MipAnalysis object
    mip_analysis_object = parse_mip_analysis(
        mip_config_raw=mip_analysis_config_dna_raw,
        qc_metrics_raw=mip_metrics_deliverables_raw,
        sample_info_raw=sample_info_dna_raw,
    )

    # THEN assert that it was successfully created
    assert isinstance(mip_analysis_object, MipAnalysis)



"""Tests for the upload genotypes api"""

from datetime import datetime
from pathlib import Path

from cg.meta.upload.gt import UploadGenotypesAPI
from cg.models.mip.mip_metrics_deliverables import MetricsDeliverables
from cg.store import models


def test_get_analysis_sex(case_qc_metrics_deliverables: Path, genotype_analysis_sex: dict):
    """Test to get the predicted sex from a MIP run using the upload genotypes API"""
    # GIVEN a UploadGenotypesAPI some qcmetrics data

    # WHEN fetching the predicted sex by the analysis
    sex: dict = UploadGenotypesAPI.analysis_sex(
        self=UploadGenotypesAPI, qc_metrics_file=case_qc_metrics_deliverables
    )

    # THEN assert that the the predicted sex per sample_id is returned
    assert sex == genotype_analysis_sex


def test_get_parsed_qc_metrics_data(case_qc_metrics_deliverables: Path):
    """Test to get the predicted sex from a MIP run using the upload genotypes API"""
    # GIVEN a UploadGenotypesAPI and the path to a qc_metrics file with case data

    # WHEN fetching the predicted sex
    metrics_object: MetricsDeliverables = UploadGenotypesAPI.get_parsed_qc_metrics_data(
        case_qc_metrics_deliverables
    )

    # THEN assert that it was successfully created
    assert isinstance(metrics_object, MetricsDeliverables)


def test_get_bcf_file(upload_genotypes_api: UploadGenotypesAPI, case_id: str, timestamp: datetime):
    """Test to get the predicted sex from a MIP run using the upload genotypes API"""
    # GIVEN a UploadGenotypesAPI populated with some data in housekeeper
    hk_version = upload_genotypes_api.hk.version(case_id, timestamp)

    # WHEN fetching the gbcf file with the api
    gbcf = upload_genotypes_api.get_bcf_file(hk_version)

    # THEN assert that the file has the correct tag
    assert "snv-gbcf" in (tag.name for tag in gbcf.tags)


def test_get_data(
    analysis_obj: models.Analysis,
    genotype_analysis_sex: dict,
    mocker,
    upload_genotypes_api: UploadGenotypesAPI,
):
    """Test to get data from the UploadGenotypesAPI"""
    # GIVEN a UploadGenotypeAPI populated with some data
    # GIVEN an analysis object with a trio

    # GIVEN analysis sex were generated and could be found
    mocker.patch.object(UploadGenotypesAPI, "analysis_sex")
    UploadGenotypesAPI.analysis_sex.return_value = genotype_analysis_sex

    # WHEN parsing the data
    result = upload_genotypes_api.data(analysis_obj=analysis_obj)

    # THEN assert that the result looks like expected
    assert len(result["samples_sex"]) == 3



"""Tests for the transfer of external data"""
import logging
from pathlib import Path
import os

from tests.cli.workflow.mip.conftest import fixture_dna_mip_context
from tests.cli.workflow.mip.conftest import fixture_mip_case_ids
from tests.cli.workflow.mip.conftest import fixture_mip_case_id
from tests.apps.mip.conftest import tb_api
from cg.meta.orders.external_data import ExternalDataAPI
from cg.models.cg_config import CGConfig
from cg.store import Store, models


def test_create_log_dir(caplog, external_data_api: ExternalDataAPI):
    """Test generating the directory for logging"""
    caplog.set_level(logging.INFO)

    # WHEN the log directory is created
    log_dir = external_data_api.create_log_dir(ticket_id=999999, dry_run=True)

    # THEN the path is not created since it is a dry run
    assert "Would have created path" in caplog.text

    # THEN the created path is
    assert str(log_dir).startswith("/another/path/999999")


def test_create_source_path(external_data_api: ExternalDataAPI):
    """Test generating the source path"""

    # WHEN the source path is created
    source_path = external_data_api.create_source_path(
        ticket_id=999999, raw_path="/path/%s", cust_sample_id="ABC123", cust_id="cust000"
    )

    # THEN the source path is
    assert source_path == "/path/cust000/999999/ABC123/"


def test_create_destination_path(external_data_api: ExternalDataAPI):
    """Test generating the destination path"""

    # WHEN the source path is created
    destination_path = external_data_api.create_destination_path(
        raw_path="/path/%s", lims_sample_id="ACC123", cust_id="cust000"
    )

    # THEN the source path is
    assert destination_path == "/path/cust000/ACC123/"


def test_download_sample(external_data_api: ExternalDataAPI, mocker):
    """Test for downloading external data via slurm"""

    # GIVEN paths needed to run rsync
    mocker.patch.object(ExternalDataAPI, "create_log_dir")
    ExternalDataAPI.create_log_dir.return_value = Path("/path/to/log")

    mocker.patch.object(ExternalDataAPI, "create_source_path")
    ExternalDataAPI.create_source_path.return_value = Path("/path/to/source")

    mocker.patch.object(ExternalDataAPI, "create_destination_path")
    ExternalDataAPI.create_destination_path.return_value = Path("/path/to/destination")

    # WHEN the destination path is created
    sbatch_number = external_data_api.download_sample(
        cust_id="cust000",
        ticket_id=123456,
        cust_sample_id="ABC123",
        lims_sample_id="ACC123",
        dry_run=True,
    )

    # THEN check that an integer was returned as sbatch number
    assert isinstance(sbatch_number, int)


def test_get_all_fastq(
    cg_context: CGConfig, external_data_directory, external_data_api: ExternalDataAPI
):
    """Test the finding of fastq.gz files in customer directories."""
    for folder in os.listdir(external_data_directory):
        # WHEN the list of file-paths is created
        files = external_data_api.get_all_fastq(
            sample_folder=str(external_data_directory) + "/" + folder
        )
        # THEN only fast.gz files are returned
        assert [tmp.endswith("fastq.gz") for tmp in files]


def test_configure_housekeeper(
    external_data_api: ExternalDataAPI,
    caplog,
    mocker,
    case_id,
    ticket_nr,
    dna_mip_context,
):
    caplog.set_level(logging.INFO)
    """Test the finding of fastq.gz files in customer directories."""

    # GIVEN a case to analyze
    mip_api = dna_mip_context.meta_apis["analysis_api"]
    cases = mip_api.status_db.query(models.Family).filter(models.Family.internal_id == case_id)

    # GIVEN a case is available for analysis
    mocker.patch.object(Store, "get_cases_from_ticket")
    Store.get_cases_from_ticket.return_value = cases

    # GIVEN a list of paths
    mocker.patch.object(ExternalDataAPI, "get_all_paths")
    ExternalDataAPI.get_all_paths.return_value = ["/path/to/something/"]

    # WHEN files are added to housekeeper
    external_data_api.configure_housekeeper(ticket_id=ticket_nr, dry_run=True)

    # THEN it is announced that files would have been added
    assert "Would have" in caplog.text



import logging
from pathlib import Path

import pytest
from cg.exc import FlowcellError
from cg.models.demultiplex.run_parameters import RunParameters


def test_get_flowcell_type_hiseq_run_parameters(hiseq_run_parameters: Path):
    # GIVEN the path to a file with hiseq run parameters
    # GIVEN a RunParameters object
    run_parameters_object = RunParameters(hiseq_run_parameters)

    # WHEN fetching the flowcell type

    # THEN assert that the flowcell type is correct
    assert run_parameters_object.flowcell_type == "hiseq"


def test_get_flowcell_type_novaseq_run_parameters(novaseq_run_parameters: Path):
    # GIVEN the path to a file with hiseq run parameters
    # GIVEN a RunParameters object
    run_parameters_object = RunParameters(novaseq_run_parameters)

    # WHEN fetching the flowcell type

    # THEN assert that the flowcell type is correct
    assert run_parameters_object.flowcell_type == "novaseq"


def test_get_unknown_type_run_parameters(unknown_run_parameters: Path, caplog):
    caplog.set_level(logging.INFO)
    # GIVEN the path to a file with a unknown flowcell type
    # GIVEN a run parameters object
    run_parameters_obj = RunParameters(unknown_run_parameters)

    # WHEN fetching the flowcell type
    with pytest.raises(FlowcellError):
        # THEN assert that an exception was raised since the flowcell type was unknown
        run_parameters_obj.flowcell_type

    assert "Unknown flowcell type" in caplog.text


def test_get_flowcell_type_when_missing(run_parameters_missing_flowcell_type: Path, caplog):
    caplog.set_level(logging.INFO)
    # GIVEN the path to a file where the field that indicates flowcell type is missing
    # GIVEN a run parameters object
    run_parameters_obj = RunParameters(run_parameters_missing_flowcell_type)

    # WHEN parsing the run parameters
    with pytest.raises(FlowcellError):
        # THEN assert that an exception was raised since the flowcell type was unknown
        run_parameters_obj.flowcell_type

    assert "Could not determine flowcell type" in caplog.text



import random
import secrets
from time import sleep
from typing import Optional

import petname
import requests
from bing_image_urls import bing_image_urls


class Avatar:
    @staticmethod
    def get_avatar_url(internal_id: str) -> Optional[bool]:
        for url in Avatar.get_avatar_urls(internal_id):
            if Avatar.is_url_image(url):
                return url

    @staticmethod
    def get_avatar_urls(internal_id: str) -> Optional[bool]:

        urls = None
        adjective, animal = Avatar._split_petname(internal_id)
        filter_array = [
            "+filterui:license-L2_L3",
            "+filterui:aspect-square",
            "+filterui:photo-transparent",
            "+filterui:imagesize-small",
        ]

        try_cnt = 0
        filters = "".join(filter_array)
        while not urls and filters:
            query = f"{adjective} {animal}"
            urls = bing_image_urls(
                query=query,
                filters=filters,
                verify_status_only=False,
            )
            if not urls:
                sleep(secrets.SystemRandom().randint(0, max(len(filter_array) - try_cnt, 1)))
                query = f"{animal}"
                urls = bing_image_urls(
                    query=query,
                    filters=filters,
                    verify_status_only=False,
                )

            try_cnt += 1
            filters = "".join(filter_array[: len(filter_array) - try_cnt])

        random.shuffle(urls)
        return urls

    @staticmethod
    def is_url_image(image_url: str) -> bool:
        from PIL import Image

        try:
            img = Image.open(requests.get(image_url, stream=True).raw)
            width, height = img.size

            if 0 < width and 0 < height:
                return True
        except Exception:
            return False

        return False

    @classmethod
    def _split_petname(cls, merged_name: str) -> (str, str):

        found_adj = ""
        sorted_adjectives = sorted(petname.adjectives, key=len, reverse=True)
        sorted_pets = sorted(petname.names, key=len, reverse=True)
        for adjective in sorted_adjectives:
            if merged_name.startswith(adjective):
                found_adj = adjective
                break
        found_pet = ""
        for pet in sorted_pets:
            if merged_name.endswith(pet):
                found_pet = pet
                break

        return found_adj, found_pet



"""Test MIP metrics deliverables"""

from cg.models.mip.mip_metrics_deliverables import (
    MetricsDeliverables,
    DuplicateReads,
    ParsedMetrics,
    GenderCheck,
    MeanInsertSize,
    MappedReads,
    get_sample_id_metric,
)


def test_instantiate_mip_metrics_deliverables(mip_metrics_deliverables_raw: dict):
    """
    Tests raw data deliverable against a pydantic MetricsDeliverables
    """
    # GIVEN a dictionary with the some metrics

    # WHEN instantiating a MetricsDeliverables object
    metrics_object = MetricsDeliverables(**mip_metrics_deliverables_raw)

    # THEN assert that it was successfully created
    assert isinstance(metrics_object, MetricsDeliverables)


def test_instantiate_mip_metrics_sample_ids(mip_metrics_deliverables_raw: dict):
    """
    Tests set sample_ids
    """
    # GIVEN a dictionary with the some metrics

    # WHEN instantiating a MetricsDeliverables object
    metrics_object = MetricsDeliverables(**mip_metrics_deliverables_raw)

    # THEN assert that sample_ids was successfully created
    assert metrics_object.sample_ids == {"an_id", "another_id"}


def test_mip_metrics_set_duplicate_reads(mip_metrics_deliverables_raw: dict):
    """
    Tests set duplicates read
    """
    # GIVEN a dictionary with the some metrics

    # WHEN instantiating a MetricsDeliverables object
    metrics_object = MetricsDeliverables(**mip_metrics_deliverables_raw)

    # THEN assert that read duplicates was set
    assert metrics_object.duplicate_reads

    duplicate_read: DuplicateReads = metrics_object.duplicate_reads.pop()

    assert isinstance(duplicate_read, DuplicateReads)

    expected_duplicate_read: dict = mip_metrics_deliverables_raw["metrics"].pop()

    # THEN assert that sample id was set
    assert duplicate_read.sample_id == expected_duplicate_read["id"]

    # THEN assert that value was set
    assert duplicate_read.value == float(expected_duplicate_read["value"]) * 100


def test_mip_metrics_set_mapped_reads(mip_metrics_deliverables_raw: dict):
    """
    Tests set mapped reads
    """
    # GIVEN a dictionary with the some metrics

    # WHEN instantiating a MetricsDeliverables object
    metrics_object = MetricsDeliverables(**mip_metrics_deliverables_raw)

    # THEN assert that mapped reads was set
    assert metrics_object.mapped_reads

    mapped_reads: MappedReads = metrics_object.mapped_reads.pop()

    # THEN assert that it was successfully created
    assert isinstance(mapped_reads, MappedReads)


def test_mip_metrics_set_mean_insert_size(mip_metrics_deliverables_raw: dict):
    """
    Tests set predicted sex
    """
    # GIVEN a dictionary with the some metrics

    # WHEN instantiating a MetricsDeliverables object
    metrics_object = MetricsDeliverables(**mip_metrics_deliverables_raw)

    # THEN assert that mean insert size was set
    assert metrics_object.mean_insert_size

    mean_insert_size: MeanInsertSize = metrics_object.mean_insert_size.pop()

    # THEN assert that it was successfully created
    assert isinstance(mean_insert_size, MeanInsertSize)


def test_mip_metrics_set_predicted_sex(mip_metrics_deliverables_raw: dict):
    """
    Tests set predicted sex
    """
    # GIVEN a dictionary with the some metrics

    # WHEN instantiating a MetricsDeliverables object
    metrics_object = MetricsDeliverables(**mip_metrics_deliverables_raw)

    # THEN assert that predicted sex was set
    assert metrics_object.predicted_sex

    predicted_sex: GenderCheck = metrics_object.predicted_sex.pop()

    # THEN assert that it was successfully created
    assert isinstance(predicted_sex, GenderCheck)


def test_instantiate_mip_metrics_set_sample_id_metrics(mip_metrics_deliverables_raw: dict):
    """
    Tests set sample_id metrics
    """
    # GIVEN a dictionary with the some metrics

    # WHEN instantiating a MetricsDeliverables object
    metrics_object = MetricsDeliverables(**mip_metrics_deliverables_raw)

    # THEN assert that sample_id metrics was set
    assert metrics_object.sample_id_metrics

    # WHEN looping through sample_id metrics
    for sample_id_metric in metrics_object.sample_id_metrics:

        # THEN assert that sample_id metrics was successfully created
        assert isinstance(sample_id_metric, ParsedMetrics)

        # THEN assert that metrics are set for sample_id
        if sample_id_metric.sample_id == "an_id":
            assert sample_id_metric.duplicate_reads == 0.0748899652117993 * 100
            assert sample_id_metric.mapped_reads == 0.9975489233589259 * 100
            assert sample_id_metric.mean_insert_size == 422.0
            assert sample_id_metric.predicted_sex == "female"


def test_get_sample_id_metric(mip_metrics_deliverables_raw: dict):
    """
    Tests get sample_id metrics
    """
    # GIVEN a dictionary with the some metrics

    # WHEN instantiating a MetricsDeliverables object
    metrics_object = MetricsDeliverables(**mip_metrics_deliverables_raw)

    # When getting an sample_id_metric
    sample_id_metric = get_sample_id_metric(
        sample_id="an_id", sample_id_metrics=metrics_object.sample_id_metrics
    )

    # THEN assert that sample_id metric for sample_id was returned
    assert sample_id_metric.sample_id == "an_id"



from datetime import datetime

import pytest
from tests.meta.report.comparison import (
    is_float,
    is_similar_values,
    is_similar_lists,
    value_exists_in,
    dict_values_exists_in,
)


@pytest.mark.parametrize("data", [0, "0", 0.1, "0.1", "1.0", "10", True, False])
def test_is_float_with_float(data):
    # GIVEN data considered float equivalent

    # WHEN testing is_float
    is_a_float = is_float(data)

    # THEN the result should be true
    assert is_a_float


@pytest.mark.parametrize("data", [None, "a", {}, (), ""])
def test_is_float_with_non_float(data):
    # GIVEN data not considered float equivalent

    # WHEN testing is_float
    is_a_float = is_float(data)

    # THEN the result should be true
    assert not is_a_float


@pytest.mark.parametrize(
    "data",
    [
        0,
        "0",
        0.1,
        "0.1",
        "1.0",
        "10",
        True,
        False,
        None,
        "",
        (),
        [],
        {},
        ["a", "b", "c"],
        {"numbers": 123},
        {"dict": {"text": "text"}},
        datetime.now(),
    ],
)
def test_is_similar_values_data_compared_to_equal(data):
    # GIVEN data

    # WHEN using two equal data items
    is_similar = is_similar_values(data, data)

    # THEN the result should be similar
    assert is_similar


@pytest.mark.parametrize(
    "data_pair",
    [(0, "0"), (0.1, "0.1"), (1.33, "1.3"), (datetime.now(), str(datetime.now().date()))],
)
def test_is_similar_values_data_compared_to_similar(data_pair):
    # GIVEN data and similar data
    data, similar_data = data_pair

    # WHEN using data and similar data
    is_similar = is_similar_values(data, similar_data)

    # THEN the result should be similar
    assert is_similar


@pytest.mark.parametrize(
    "data_pair",
    [(0, "1"), (0.1, None), (None, "1.3"), (str(datetime.now().date()), datetime.now())],
)
def test_is_similar_values_data_compared_to_non_similar(data_pair):
    # GIVEN data and similar data
    data, non_similar_data = data_pair

    # WHEN using data and non similar data
    is_similar = is_similar_values(data, non_similar_data)

    # THEN the result should not be similar
    assert not is_similar


@pytest.mark.parametrize(
    "data",
    [None, "abc", [1, 2, 3], ["a", "b", "c"], ["a", 1, "1", None, [], {}, (), {}, datetime.now()]],
)
def test_is_similar_lists_compared_to_equal(data):
    # GIVEN data

    # WHEN using two equal data items
    is_similar = is_similar_lists(data, data)

    # THEN the result should be similar
    assert is_similar


@pytest.mark.parametrize("data", ["some_text", 1.3, 123, datetime.now(), True])
def test_value_exists_in_text(data):
    # GIVEN data and a text containing that data
    data_in_text = "this is a text " + str(data) + " and then some more text"

    # WHEN running function
    exists = value_exists_in(data, data_in_text)

    # THEN the result should be True
    assert exists


@pytest.mark.parametrize("data", ["some_text", 1.3, 123, datetime.now()])
def test_value_exists_not_in_text(data):
    # GIVEN data and a text not missing that data
    data_not_in_text = "this is a text and then some more text"

    # WHEN running function
    exists = value_exists_in(data, data_not_in_text)

    # THEN the result should be True
    assert not exists


@pytest.mark.parametrize(
    "data_pair",
    [
        (
            {"text": "some_text", "dict": {"number": 1.3, "list": ["abc", 123, True]}},
            "Report containing some_text the number 1.3, the list abc and 123",
        )
    ],
)
def test_dict_values_exists_in_text(data_pair):
    # GIVEN data and a text containing that data
    data, data_in_text = data_pair

    # WHEN running function
    exists = dict_values_exists_in(data, data_in_text)

    # THEN the result should be True
    assert exists


@pytest.mark.parametrize(
    "data_pair",
    [
        (
            {
                "text": "some_text",
                "dict": {"number": 1.3, "list": ["abc", 123, datetime.now(), True]},
            },
            "Report not containing data",
        )
    ],
)
def test_dict_values_not_existing_in_text(data_pair):
    # GIVEN data and a text not containing that data
    data, data_in_text = data_pair

    # WHEN running function
    exists = dict_values_exists_in(data, data_in_text)

    # THEN the result should be False
    assert not exists



"""Test module cg.cli.upload.nipt"""
import datetime
import json
import logging

from cg.cli.upload.nipt.base import nipt_upload_case, nipt_upload_all
from cg.meta.upload.nipt import NiptUploadAPI
from cg.meta.upload.nipt.models import StatinaUploadFiles
from cgmodels.cg.constants import Pipeline
from click.testing import CliRunner

from cg.models.cg_config import CGConfig

NIPT_CASE_SUCCESS = "*** NIPT UPLOAD START ***"
NIPT_ALL_SUCCESS = "*** NIPT UPLOAD ALL START ***"
NIPT_STATINA_SUCCESS = "*** Statina UPLOAD START ***"
NIPT_FTP_SUCCESS = "*** NIPT FTP UPLOAD START ***"


class MockStatinaUploadFiles:
    def json(self, *args, **kwargs):
        return ""


def test_nipt_statina_upload_case(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, helpers, mocker
):
    """Tests CLI command to upload a single case"""

    # GIVEN a specified NIPT case that has its analysis stored but is not yet uploaded
    caplog.set_level(logging.DEBUG)

    analysis_obj = helpers.add_analysis(store=upload_context.status_db)
    case_id = analysis_obj.family.internal_id
    assert not analysis_obj.upload_started_at
    assert not analysis_obj.uploaded_at

    # WHEN uploading of a specified NIPT case
    mocker.patch.object(NiptUploadAPI, "get_statina_files", return_value=MockStatinaUploadFiles())
    mocker.patch.object(NiptUploadAPI, "upload_to_statina_database")
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    result = cli_runner.invoke(
        nipt_upload_case, [case_id], obj=upload_context, catch_exceptions=False
    )

    # THEN both the nipt ftp and statina upload should start
    assert NIPT_CASE_SUCCESS in caplog.text
    assert NIPT_STATINA_SUCCESS in caplog.text
    assert NIPT_FTP_SUCCESS in caplog.text

    # THEN set analysis.upload_started_at in the database
    assert analysis_obj.upload_started_at

    # THEN set analysis.uploaded_at in the database
    assert analysis_obj.uploaded_at

    # THEN exit without errors
    assert result.exit_code == 0


def test_nipt_statina_upload_case_dry_run(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, helpers, mocker
):
    """Tests CLI command to upload a single case"""

    # GIVEN a specified NIPT case that has its analysis stored but is not yet uploaded
    caplog.set_level(logging.DEBUG)

    analysis_obj = helpers.add_analysis(store=upload_context.status_db)
    case_id = analysis_obj.family.internal_id
    assert not analysis_obj.upload_started_at
    assert not analysis_obj.uploaded_at

    # WHEN uploading a specified NIPT case with dry-run flag set
    mocker.patch.object(NiptUploadAPI, "get_statina_files", return_value=MockStatinaUploadFiles())
    mocker.patch.object(NiptUploadAPI, "upload_to_statina_database")
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    result = cli_runner.invoke(nipt_upload_case, [case_id, "--dry-run"], obj=upload_context)

    # THEN both the nipt ftp and statina upload should start
    assert NIPT_CASE_SUCCESS in caplog.text
    assert NIPT_STATINA_SUCCESS in caplog.text
    assert NIPT_FTP_SUCCESS in caplog.text

    # THEN analysis.upload_started_at should not be set in the database
    assert analysis_obj.upload_started_at is None

    # THEN analysis.uploaded_at should not be set in the database
    assert analysis_obj.uploaded_at is None

    # THEN exit without errors
    assert result.exit_code == 0


def test_nipt_statina_upload_auto(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, helpers, mocker
):
    """Tests CLI command to upload a single case"""

    # GIVEN a case ready for upload
    caplog.set_level(logging.DEBUG)

    analysis_obj = helpers.add_analysis(
        store=upload_context.status_db,
        completed_at=datetime.datetime.now(),
        pipeline=Pipeline.FLUFFY,
    )
    assert analysis_obj.completed_at
    assert not analysis_obj.uploaded_at

    # WHEN uploading all NIPT cases
    mocker.patch.object(NiptUploadAPI, "get_statina_files", return_value=MockStatinaUploadFiles())
    mocker.patch.object(NiptUploadAPI, "upload_to_statina_database")
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    result = cli_runner.invoke(nipt_upload_all, [], obj=upload_context)

    # THEN both the nipt ftp and statina upload should start
    assert NIPT_ALL_SUCCESS in caplog.text
    assert NIPT_STATINA_SUCCESS in caplog.text
    assert NIPT_FTP_SUCCESS in caplog.text

    # THEN set analysis.upload_started_at in the database
    assert analysis_obj.upload_started_at

    # THEN set analysis.uploaded_at in the database
    assert analysis_obj.uploaded_at

    # THEN exit without errors
    assert result.exit_code == 0


def test_nipt_statina_upload_auto_dry_run(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, helpers, mocker
):
    """Tests CLI command to upload a single case"""

    # GIVEN a case ready for upload
    caplog.set_level(logging.DEBUG)

    analysis_obj = helpers.add_analysis(
        store=upload_context.status_db,
        completed_at=datetime.datetime.now(),
        pipeline=Pipeline.FLUFFY,
    )
    assert analysis_obj.completed_at
    assert not analysis_obj.uploaded_at

    # WHEN uploading all NIPT cases
    mocker.patch.object(NiptUploadAPI, "get_statina_files", return_value=MockStatinaUploadFiles())
    mocker.patch.object(NiptUploadAPI, "upload_to_statina_database")
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    result = cli_runner.invoke(nipt_upload_all, ["--dry-run"], obj=upload_context)

    # THEN both the nipt ftp and statina upload should start
    assert NIPT_ALL_SUCCESS in caplog.text
    assert NIPT_STATINA_SUCCESS in caplog.text
    assert NIPT_FTP_SUCCESS in caplog.text

    # THEN analysis.upload_started_at should not be set in the database
    assert analysis_obj.upload_started_at is None

    # THEN analysis.uploaded_at should not be set in the database
    assert analysis_obj.uploaded_at is None

    # THEN exit without errors
    assert result.exit_code == 0



from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel


class StatsSample(BaseModel):
    name: str
    reads: int = 0
    fastqs: List[str] = []


class StatsFlowcell(BaseModel):
    name: str
    sequencer: Optional[str]
    sequencer_type: str
    date: datetime
    samples: List[StatsSample]



from cg.apps.cgstats.crud import create, find
from cg.apps.cgstats.stats import StatsAPI
from cg.models.demultiplex.demux_results import DemuxResults

from cg.apps.cgstats.db import models as stats_models


def test_create_support_parameters(stats_api: StatsAPI, demux_results: DemuxResults):
    # GIVEN a cg stats api without any support parameter for the demux result
    assert not find.get_support_parameters_id(demux_results=demux_results)

    # WHEN creating the new support parameters
    create.create_support_parameters(manager=stats_api, demux_results=demux_results)

    # THEN assert that the parameters was created
    support_parameters_id = find.get_support_parameters_id(demux_results=demux_results)
    assert isinstance(support_parameters_id, int)


def test_create_data_source(stats_api: StatsAPI, demux_results: DemuxResults):
    # GIVEN a api with some support parameters
    support_parameters: stats_models.Supportparams = create.create_support_parameters(
        manager=stats_api, demux_results=demux_results
    )
    # GIVEN that there are no data source for the given run
    assert not find.get_datasource_id(demux_results=demux_results)

    # WHEN creating a new datasource
    create.create_datasource(
        manager=stats_api,
        demux_results=demux_results,
        support_parameters_id=support_parameters.supportparams_id,
    )

    # THEN assert that the datasource exists
    assert find.get_datasource_id(demux_results=demux_results)


def test_create_flowcell(stats_api: StatsAPI, demux_results: DemuxResults):
    # GIVEN a api without a flowcell object
    assert not find.get_flowcell_id(flowcell_name=demux_results.flowcell.flowcell_id)

    # WHEN creating a new flowcell
    create.create_flowcell(manager=stats_api, demux_results=demux_results)

    # THEN assert that the flowcell was created
    assert find.get_flowcell_id(flowcell_name=demux_results.flowcell.flowcell_id)


def test_create_demux(stats_api: StatsAPI, demux_results: DemuxResults):
    # GIVEN a database with a flowcell and a data source
    support_parameters: stats_models.Supportparams = create.create_support_parameters(
        manager=stats_api, demux_results=demux_results
    )
    flowcell: stats_models.Flowcell = create.create_flowcell(
        manager=stats_api, demux_results=demux_results
    )
    data_source: stats_models.Datasource = create.create_datasource(
        manager=stats_api,
        demux_results=demux_results,
        support_parameters_id=support_parameters.supportparams_id,
    )
    # GIVEN that there is not demux object in the database
    assert not find.get_demux_id(flowcell_object_id=flowcell.flowcell_id)

    # WHEN creating a demux object
    create.create_demux(
        manager=stats_api, flowcell_id=flowcell.flowcell_id, datasource_id=data_source.datasource_id
    )

    # THEN assert that a demux object was created
    assert find.get_demux_id(flowcell_object_id=flowcell.flowcell_id)



"""Tests for running the deploy command"""

import logging
from pathlib import Path

from cg.cli.deploy.base import deploy
from cg.models.cg_config import CGConfig, CommonAppConfig
from click.testing import CliRunner


def test_running_deploy_fluffy(shipping_context: CGConfig, fluffy_config: Path, caplog):
    """Test to deploy fluffy with CG"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with some shipping configs

    shipping_context.fluffy = CommonAppConfig(binary_path="echo", deploy_config=str(fluffy_config))
    # GIVEN a cli runner
    runner = CliRunner()

    # WHEN running the deploy fluffy command in dry run mode
    result = runner.invoke(deploy, ["--dry-run", "fluffy"], obj=shipping_context)

    # THEN assert that the command exits without problems
    assert result.exit_code == 0


def test_running_deploy_genotype(shipping_context: CGConfig, caplog):
    """Test to deploy genotype with CG"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with some shipping configs

    # GIVEN a cli runner
    runner = CliRunner()

    # WHEN running the deploy genotype command in dry run mode
    result = runner.invoke(deploy, ["--dry-run", "genotype"], obj=shipping_context)

    # THEN assert that the command exits without problems
    assert result.exit_code == 0


def test_running_deploy_hermes(shipping_context: CGConfig, hermes_config: Path, caplog):
    """Test to deploy hermes with CG"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with some shipping configs

    shipping_context.hermes = CommonAppConfig(binary_path="echo", deploy_config=str(hermes_config))
    # GIVEN a cli runner
    runner = CliRunner()

    # WHEN running the deploy hermes command in dry run mode
    result = runner.invoke(deploy, ["--dry-run", "hermes"], obj=shipping_context)

    # THEN assert that the command exits without problems
    assert result.exit_code == 0


def test_running_deploy_loqusdb(shipping_context: CGConfig, caplog):
    """Test to deploy loqusdb with CG"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with some shipping configs

    # GIVEN a cli runner
    runner = CliRunner()

    # WHEN running the deploy loqusdb command in dry run mode
    result = runner.invoke(deploy, ["--dry-run", "loqusdb"], obj=shipping_context)

    # THEN assert that the command exits without problems
    assert result.exit_code == 0


def test_running_deploy_scout(shipping_context: CGConfig, scout_config: Path, caplog):
    """Test to deploy scout with CG"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with some shipping configs

    shipping_context.scout = CommonAppConfig(binary_path="echo", deploy_config=str(scout_config))
    # GIVEN a cli runner
    runner = CliRunner()

    # WHEN running the deploy scout command in dry run mode
    result = runner.invoke(deploy, ["--dry-run", "scout"], obj=shipping_context)

    # THEN assert that the command exits without problems
    assert result.exit_code == 0


def test_running_deploy_shipping(shipping_context: CGConfig, base_context: CGConfig, caplog):
    """Test to deploy shipping with CG"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with some shipping configs

    # GIVEN a cli runner
    runner = CliRunner()

    # WHEN running the deploy shipping command in dry run mode
    result = runner.invoke(deploy, ["--dry-run", "shipping"], obj=shipping_context)

    # THEN assert that the command exits without problems
    assert result.exit_code == 0



"""
    Tests for AvatarAPI
"""
import pytest
from cg.apps.avatar.api import Avatar


def test_split_petname():
    """Test to instantiate a avatar api"""
    # GIVEN a merged petname
    merged_petname = "simpleox"

    # WHEN splitting the petname
    split_adjective, split_pet = Avatar._split_petname(merged_petname)

    # THEN the parts of the petname should be the same as the original merged one
    assert split_adjective + split_pet == merged_petname


@pytest.mark.parametrize("adjective", ["apt", "adapted"])
def test_split_petname_ambigous_adjective(adjective):
    """Test to instantiate a avatar api"""
    # GIVEN a merged petname from an ambigous adjective and straightforward pet
    pet = "husky"
    merged_petname = adjective + pet

    # WHEN splitting the petname
    split_petname = Avatar._split_petname(merged_petname)

    # THEN the parts of the petname should be the same as the original merged one
    assert split_petname[0] + split_petname[1] == merged_petname


@pytest.mark.parametrize("pet", ["cat", "cattle", "wildcat"])
def test_split_petname_ambigous_pet(pet):
    """Test to instantiate a avatar api"""
    # GIVEN a merged petname from an ambigous pet and straightforward adjective
    adjective = "strong"
    merged_petname = adjective + pet

    # WHEN splitting the petname
    split_petname = Avatar._split_petname(merged_petname)

    # THEN the parts of the petname should be the same as the original merged one
    assert split_petname[0] + split_petname[1] == merged_petname



import json

import pytest
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.osticket import MockOsTicket

from cg.meta.orders import OrdersAPI, OrderType
from cg.meta.orders.status import StatusHandler
from cg.meta.orders.ticket_handler import TicketHandler
from cg.models.orders.order import OrderIn
from cg.store import Store
from pathlib import Path


@pytest.fixture
def all_orders_to_submit(
    balsamic_order_to_submit,
    external_order_to_submit,
    fastq_order_to_submit,
    metagenome_order_to_submit,
    microbial_order_to_submit,
    mip_order_to_submit,
    mip_rna_order_to_submit,
    rml_order_to_submit,
    sarscov2_order_to_submit,
):
    return {
        OrderType.BALSAMIC: OrderIn.parse_obj(balsamic_order_to_submit),
        OrderType.EXTERNAL: OrderIn.parse_obj(external_order_to_submit),
        OrderType.FASTQ: OrderIn.parse_obj(fastq_order_to_submit),
        OrderType.METAGENOME: OrderIn.parse_obj(metagenome_order_to_submit),
        OrderType.MICROSALT: OrderIn.parse_obj(microbial_order_to_submit),
        OrderType.MIP_DNA: OrderIn.parse_obj(mip_order_to_submit),
        OrderType.MIP_RNA: OrderIn.parse_obj(mip_rna_order_to_submit),
        OrderType.RML: OrderIn.parse_obj(rml_order_to_submit),
        OrderType.SARS_COV_2: OrderIn.parse_obj(sarscov2_order_to_submit),
    }


@pytest.fixture
def rml_status_data(rml_order_to_submit):
    """Parse rml order example."""
    return StatusHandler.pools_to_status(rml_order_to_submit)


@pytest.fixture
def fastq_status_data(fastq_order_to_submit):
    """Parse fastq order example."""
    return StatusHandler.samples_to_status(fastq_order_to_submit)


@pytest.fixture
def mip_status_data(mip_order_to_submit):
    """Parse scout order example."""
    return StatusHandler.cases_to_status(mip_order_to_submit)


@pytest.fixture
def mip_rna_status_data(mip_rna_order_to_submit):
    """Parse rna order example."""
    return StatusHandler.cases_to_status(mip_rna_order_to_submit)


@pytest.fixture
def external_status_data(external_order_to_submit):
    """Parse external order example."""
    return StatusHandler.cases_to_status(external_order_to_submit)


@pytest.fixture
def microbial_status_data(microbial_order_to_submit):
    """Parse microbial order example."""
    return StatusHandler.microbial_samples_to_status(microbial_order_to_submit)


@pytest.fixture
def metagenome_status_data(metagenome_order_to_submit):
    """Parse metagenome order example."""
    return StatusHandler.samples_to_status(metagenome_order_to_submit)


@pytest.fixture
def balsamic_status_data(balsamic_order_to_submit):
    """Parse cancer order example."""
    return StatusHandler.cases_to_status(balsamic_order_to_submit)


@pytest.fixture(scope="function")
def orders_api(base_store, osticket: MockOsTicket, lims_api: MockLimsAPI):
    return OrdersAPI(lims=lims_api, status=base_store, osticket=osticket)


@pytest.fixture(name="ticket_handler")
def fixture_ticket_handler(store: Store, osticket: MockOsTicket) -> TicketHandler:
    return TicketHandler(status_db=store, osticket_api=osticket)


@pytest.fixture(name="external_data_directory")
def external_data_directory(tmpdir_factory) -> Path:
    """Fixture that returns a customer folder with fastq.gz files in sample-directories."""
    cust_folder = tmpdir_factory.mktemp("cust000")
    Path(cust_folder, "sample1").mkdir(exist_ok=True, parents=True)
    Path(cust_folder, "sample2").mkdir(exist_ok=True, parents=True)
    Path(cust_folder, "sample1", "sample1_fastq_1.fastq.gz").touch(exist_ok=True)
    Path(cust_folder, "sample1", "sample1_fastq_2.fastq.gz").touch(exist_ok=True)
    Path(cust_folder, "sample1", "sample1_fastq_2.fastq.gz.md5").touch(exist_ok=True)
    Path(cust_folder, "sample2", "sample2_fastq_1.fastq.gz").touch(exist_ok=True)
    Path(cust_folder, "sample2", "sample2_fastq_2.fastq.gz").touch(exist_ok=True)
    Path(cust_folder, "sample2", "sample1_fastq_2.fastq.gz.md5").touch(exist_ok=True)
    return Path(cust_folder)



from typing import Optional

from cg.apps.tb.models import TrailblazerAnalysis


class MockTB:
    def __init__(self):
        self.analyses_response = []
        self.get_latest_analysis_response = {}

    def is_latest_analysis_ongoing(self, *args, **kwargs) -> bool:
        return False

    def add_pending_analysis(self, *args, **kwargs) -> None:
        return None

    def mark_analyses_deleted(self, *args, **kwargs) -> None:
        return None

    def add_commit(self, *args, **kwargs) -> None:
        return None

    def has_latest_analysis_started(self, *args, **kwargs) -> bool:
        return False

    def analyses(self, *args, **kwargs) -> list:
        return self.analyses_response

    def get_latest_analysis(self, case_id: str) -> Optional[TrailblazerAnalysis]:
        return self.get_latest_analysis_response.get(case_id)

    def ensure_analyses_response(self, analyses_list: list) -> None:
        self.analyses_response = [
            TrailblazerAnalysis.parse_obj(analysis) for analysis in analyses_list
        ]

    def ensure_get_latest_analysis_response(self, analysis_dict: dict) -> None:
        self.get_latest_analysis_response[analysis_dict["family"]] = TrailblazerAnalysis.parse_obj(
            analysis_dict
        )



import pytest
from cg.apps.pdc import PdcApi
from cg.cli.backup import MAX_FLOWCELLS_ON_DISK
from cg.meta.backup import BackupApi
from cg.models.cg_config import CGConfig


@pytest.fixture(name="backup_context")
def fixture_backup_context(cg_context: CGConfig) -> CGConfig:
    cg_context.meta_apis["backup_api"] = BackupApi(
        status=cg_context.status_db,
        pdc_api=PdcApi(),
        max_flowcells_on_disk=cg_context.max_flowcells or MAX_FLOWCELLS_ON_DISK,
        root_dir=cg_context.backup.root.dict(),
    )
    return cg_context



from typing import Optional

from sqlalchemy import Column, ForeignKey, orm, types
from sqlalchemy.orm.exc import NoResultFound

from .backup_tape import Backuptape
from .base import Model


class Backup(Model):

    runname = Column(types.String(255), primary_key=True)
    startdate = Column(types.Date, nullable=False)
    nas = Column(types.String(255))
    nasdir = Column(types.String(255))
    starttonas = Column(types.DateTime)
    endtonas = Column(types.DateTime)
    preproc = Column(types.String(255))
    preprocdir = Column(types.String(255))
    startpreproc = Column(types.DateTime)
    endpreproc = Column(types.DateTime)
    frompreproc = Column(types.DateTime)
    analysis = Column(types.String(255))
    analysisdir = Column(types.String(255))
    toanalysis = Column(types.DateTime)
    fromanalysis = Column(types.DateTime)
    inbackupdir = Column(types.Integer)
    backuptape_id = Column(ForeignKey("backuptape.backuptape_id"), nullable=False)
    backupdone = Column(types.DateTime)
    md5done = Column(types.DateTime)

    tape = orm.relationship("Backuptape", backref=orm.backref("backup"))

    @staticmethod
    def exists(runname: str, tapedir: Optional[str] = None) -> Optional[str]:
        """Check if run is already backed up. Optionally: checks if run is
        on certain tape

        runname (str): e.g. 151117_D00410_0187_AHWYGMADXX
        tapedir (str): the name of the tape, e.g. tape036_037
        """
        try:
            if tapedir is not None:
                backup: Backup = (
                    Backup.query.outerjoin(Backuptape)
                    .filter_by(runname=runname)
                    .filter(Backuptape.tapedir == tapedir)
                    .one()
                )
            else:
                backup: Backup = Backup.query.filter_by(runname=runname).one()
            return backup.runname
        except NoResultFound:
            return None



from pathlib import Path

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.housekeeper.models import InputBundle
from cg.constants import Pipeline
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.workflow.mip_rna import MipRNAAnalysisAPI
from cg.models.cg_config import CGConfig
from tests.store_helpers import StoreHelpers


@pytest.fixture(name="mip_case_id")
def fixture_mip_case_id() -> str:
    return "yellowhog"


@pytest.fixture(name="mip_case_id_non_existing")
def fixture_mip_case_id_non_existing() -> str:
    return "this_case_id_does_not_exist"


@pytest.fixture(name="mip_case_ids")
def fixture_mip_case_ids(mip_case_id: str) -> dict:
    """Dictionary of case ids, connected samples, their name and if they should fail (textbook or not)"""

    return {
        mip_case_id: {
            "textbook": True,
            "name": "F0000001",
            "internal_id": "ACC00000",
        },
        "purplesnail": {
            "textbook": False,
            "name": "F0000003",
            "internal_id": "ACC00000",
        },
        "bluezebra": {
            "textbook": True,
            "name": "F0000002",
            "internal_id": "ACC00000",
        },
    }


@pytest.fixture(name="mip_case_dirs")
def fixture_mip_case_dirs(mip_case_ids: dict, project_dir: Path) -> dict:
    """Create case directories and return a dictionary of the tmpdir paths"""

    mip_case_dirs = {}

    for case in mip_case_ids:
        case_dir = Path(project_dir / "cases" / case)
        case_dir.mkdir(exist_ok=True, parents=True)
        mip_case_dirs[case] = case_dir

    return mip_case_dirs


@pytest.fixture(name="mip_hermes_dna_deliverables_response_data")
def fixture_mip_dna_hermes_deliverables_response_data(
    create_multiqc_html_file,
    create_multiqc_json_file,
    mip_case_id,
    timestamp_yesterday,
) -> InputBundle:
    return InputBundle(
        **{
            "files": [
                {
                    "path": create_multiqc_json_file.as_posix(),
                    "tags": ["multiqc-json", mip_case_id, "mip-dna"],
                },
                {
                    "path": create_multiqc_html_file.as_posix(),
                    "tags": ["multiqc-html", mip_case_id, "mip-dna"],
                },
            ],
            "created": timestamp_yesterday,
            "name": mip_case_id,
        }
    )


@pytest.fixture(name="rna_mip_context")
def fixture_rna_mip_context(
    cg_context: CGConfig,
    analysis_family_single_case: dict,
    helpers: StoreHelpers,
    apptag_rna: str,
    case_id: str,
    housekeeper_api: HousekeeperAPI,
) -> CGConfig:
    cg_context.housekeeper_api_ = housekeeper_api
    analysis_family_single_case["data_analysis"] = str(Pipeline.MIP_RNA)
    if not cg_context.status_db.family(case_id):
        helpers.ensure_case_from_dict(
            cg_context.status_db, case_info=analysis_family_single_case, app_tag=apptag_rna
        )
    cg_context.meta_apis["analysis_api"] = MipRNAAnalysisAPI(cg_context)
    return cg_context


@pytest.fixture(name="dna_mip_context")
def fixture_dna_mip_context(
    cg_context: CGConfig,
    helpers: StoreHelpers,
    mip_case_ids: dict,
    real_housekeeper_api: HousekeeperAPI,
    tb_api,
) -> CGConfig:
    _store = cg_context.status_db
    cg_context.housekeeper_api_ = real_housekeeper_api
    cg_context.trailblazer_api_ = tb_api
    mip_analysis_api = MipDNAAnalysisAPI(config=cg_context)

    # Add apptag to db
    helpers.ensure_application_version(store=_store, application_tag="WGSA", application_type="wgs")

    # Add sample, cases and relationships to db

    for case_id in mip_case_ids:
        if not _store.family(case_id):
            case_obj = helpers.add_case(
                store=_store,
                internal_id=case_id,
                case_id=mip_case_ids[case_id]["name"],
            )
            sample = helpers.add_sample(
                store=_store,
                sample=mip_case_ids[case_id]["internal_id"],
                data_analysis=Pipeline.MIP_DNA,
                customer_name="cust000",
                application_tag="WGSA",
                application_type="wgs",
                gender="unknown",
            )
            helpers.add_relationship(store=_store, sample=sample, case=case_obj, status="affected")
    cg_context.meta_apis["analysis_api"] = mip_analysis_api
    return cg_context



"""Tests for coverage meta API"""

import datetime

from cg.apps.coverage.api import ChanjoAPI
from cg.meta.upload.coverage import UploadCoverageApi


class MockAnalysis:
    """Mock Analysis object"""

    def __init__(self, case_obj):
        self.case_obj = case_obj

    @property
    def started_at(self):
        """mock started_at date"""
        return str(datetime.datetime.today())

    @property
    def family(self):
        """mock case"""
        return self.case_obj


def test_data(coverage_upload_api, analysis_store, case_id):
    """test getting data for chanjo"""
    # GIVEN a coverage api and an analysis object
    coverage_api = coverage_upload_api
    case_name = case_id
    case_obj = analysis_store.family(case_name)
    analysis_obj = MockAnalysis(case_obj=case_obj)

    # WHEN using the data method
    results = coverage_api.data(analysis_obj=analysis_obj)

    # THEN this returns the data needed to upload samples to chanjo
    assert results["family"] == case_name
    for sample in results["samples"]:
        assert set(sample.keys()) == set(["coverage", "sample", "sample_name"])


def test_upload(chanjo_config_dict, populated_housekeeper_api, analysis_store, mocker, case_id):
    """test uploading with chanjo"""
    # GIVEN a coverage api and a data dictionary
    mock_upload = mocker.patch.object(ChanjoAPI, "upload")
    mock_sample = mocker.patch.object(ChanjoAPI, "sample")
    mock_remove = mocker.patch.object(ChanjoAPI, "delete_sample")
    hk_api = populated_housekeeper_api
    chanjo_api = ChanjoAPI(config=chanjo_config_dict)
    coverage_api = UploadCoverageApi(status_api=None, hk_api=hk_api, chanjo_api=chanjo_api)
    family_name = case_id
    case_obj = analysis_store.family(family_name)
    analysis_obj = MockAnalysis(case_obj=case_obj)
    data = coverage_api.data(analysis_obj=analysis_obj)

    # WHEN uploading samples in data dictionary
    coverage_api.upload(data=data, replace=True)

    # THEN methods sample, and upload should each have been called three times
    assert mock_upload.call_count == len(data["samples"])
    assert mock_sample.call_count == len(data["samples"])



""" Module to group PDC related commands """

import logging
import subprocess

LOG = logging.getLogger(__name__)

SERVER = "hasta"


class PdcApi:
    """Group PDC related commands"""

    @classmethod
    def retrieve_flowcell(
        cls, flowcell_id: str, sequencer_type: str, root_dir: dict, dry: bool = False
    ) -> None:
        """Fetch a flowcell back from the backup solution."""
        path = root_dir[sequencer_type]
        bash_command = f"bash SCRIPTS/retrieve_run_nas.bash {flowcell_id} {SERVER} {path}"
        command = ["ssh", "nas-9.scilifelab.se", bash_command]
        LOG.info(" ".join(command))
        if not dry:
            subprocess.check_call(command)



"""Module for building the rsync command to send files to customer inbox on caesar"""
import datetime as dt
import glob
import logging
import yaml
from pathlib import Path
from typing import List, Dict, Iterable

from cg.apps.slurm.slurm_api import SlurmAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants.priority import SlurmQos
from cg.exc import CgError
from cg.meta.meta import MetaAPI
from cg.meta.rsync.sbatch import RSYNC_COMMAND, ERROR_RSYNC_FUNCTION, COVID_RSYNC
from cg.models.cg_config import CGConfig
from cg.models.slurm.sbatch import Sbatch
from cg.store import models
from cg.constants import Pipeline
from cg.constants.priority import SLURM_ACCOUNT_TO_QOS

LOG = logging.getLogger(__name__)


class RsyncAPI(MetaAPI):
    def __init__(self, config: CGConfig):
        super().__init__(config)
        self.delivery_path: str = config.delivery_path
        self.destination_path: str = config.data_delivery.destination_path
        self.covid_destination_path: str = config.data_delivery.covid_destination_path
        self.covid_report_path: str = config.data_delivery.covid_report_path
        self.base_path: Path = Path(config.data_delivery.base_path)
        self.account: str = config.data_delivery.account
        self.log_dir: Path = Path(config.data_delivery.base_path)
        self.mail_user: str = config.data_delivery.mail_user
        self.priority: str = SlurmQos.LOW
        self.pipeline: str = Pipeline.RSYNC

    @property
    def trailblazer_config_path(self) -> Path:
        """Return Path to trailblazer config"""
        return self.log_dir / "slurm_job_ids.yaml"

    @property
    def rsync_processes(self) -> Iterable[Path]:
        """Yield existing rsync processes"""
        yield from self.base_path.iterdir()

    @staticmethod
    def format_covid_destination_path(covid_destination_path: str, customer_id: str) -> str:
        """Return destination path of covid report"""
        return covid_destination_path % customer_id

    @staticmethod
    def get_trailblazer_config(slurm_job_id: int) -> Dict[str, List[str]]:
        """Return dictionary of slurm job IDs"""
        return {"jobs": [str(slurm_job_id)]}

    @staticmethod
    def write_trailblazer_config(content: dict, config_path: Path) -> None:
        """Write slurm job IDs to a .YAML file used as the trailblazer config"""
        LOG.info(f"Writing slurm jobs to {config_path.as_posix()}")
        with config_path.open("w") as yaml_file:
            yaml.safe_dump(content, yaml_file, indent=4, explicit_start=True)

    @staticmethod
    def process_ready_to_clean(before: dt.datetime, process: Path) -> bool:
        """Return True if analysis is old enough to be cleaned"""

        ctime: dt.datetime = dt.datetime.fromtimestamp(process.stat().st_ctime)

        return before > ctime

    def set_log_dir(self, ticket_id: int) -> None:
        if self.log_dir.as_posix() == self.base_path.as_posix():
            timestamp = dt.datetime.now()
            timestamp_str = timestamp.strftime("%y%m%d_%H_%M_%S_%f")
            folder_name = Path("_".join([str(ticket_id), timestamp_str]))
            LOG.info(f"Setting log dir to: {self.base_path / folder_name}")
            self.log_dir: Path = self.base_path / folder_name

    def get_all_cases_from_ticket(self, ticket_id: int) -> List[models.Family]:
        cases: List[models.Family] = self.status_db.get_cases_from_ticket(ticket_id=ticket_id).all()
        return cases

    def get_source_and_destination_paths(self, ticket_id: int) -> Dict[str, str]:
        cases: List[models.Family] = self.get_all_cases_from_ticket(ticket_id=ticket_id)
        source_and_destination_paths = {}
        if not cases:
            LOG.warning("Could not find any cases for ticket_id %s", ticket_id)
            raise CgError()
        customer_id: str = cases[0].customer.internal_id
        source_and_destination_paths["delivery_source_path"]: Path = (
            Path(self.delivery_path, customer_id, "inbox", str(ticket_id)).as_posix() + "/"
        )
        source_and_destination_paths["rsync_destination_path"]: Path = (
            Path(self.destination_path, customer_id, "inbox", str(ticket_id)).as_posix() + "/"
        )
        return source_and_destination_paths

    def add_to_trailblazer_api(
        self, tb_api: TrailblazerAPI, slurm_job_id: int, ticket_id: int, dry_run: bool
    ) -> None:
        """Add rsync process to trailblazer"""
        if dry_run:
            return
        self.write_trailblazer_config(
            content=self.get_trailblazer_config(slurm_job_id),
            config_path=self.trailblazer_config_path,
        )
        tb_api.add_pending_analysis(
            case_id=str(ticket_id),
            analysis_type="other",
            config_path=self.trailblazer_config_path.as_posix(),
            out_dir=self.log_dir.as_posix(),
            priority=self.priority,
            email=self.mail_user,
            data_analysis=Pipeline.RSYNC,
        )

    def format_covid_report_path(self, case: models.Family, ticket_id: int) -> str:
        """Return a formatted of covid report path"""
        covid_report_options: List[str] = glob.glob(
            self.covid_report_path % (str(case.internal_id), ticket_id)
        )
        if not covid_report_options:
            LOG.error(
                f"No report file could be found with path"
                f" {self.covid_report_path % (str(case.internal_id), ticket_id)}!"
            )
            raise CgError()
        return covid_report_options[0]

    def create_log_dir(self, dry_run: bool) -> None:
        """Create log dir"""
        log_dir: Path = self.log_dir
        LOG.info("Creating folder: %s", log_dir)
        if log_dir.exists():
            LOG.warning("Could not create %s, this folder already exist", log_dir)
        elif dry_run:
            LOG.info("Would have created path %s, but this is a dry run", log_dir)
        else:
            log_dir.mkdir(parents=True, exist_ok=True)

    def run_rsync_on_slurm(self, ticket_id: int, dry_run: bool) -> int:
        self.set_log_dir(ticket_id=ticket_id)
        self.create_log_dir(dry_run=dry_run)
        source_and_destination_paths: Dict[str, str] = self.get_source_and_destination_paths(
            ticket_id=ticket_id
        )
        cases: List[models.Family] = self.get_all_cases_from_ticket(ticket_id=ticket_id)
        customer_id: str = cases[0].customer.internal_id
        if cases[0].data_analysis == Pipeline.SARS_COV_2:
            LOG.info("Delivering report for SARS-COV-2 analysis")
            commands = COVID_RSYNC.format(
                source_path=source_and_destination_paths["delivery_source_path"],
                destination_path=source_and_destination_paths["rsync_destination_path"],
                covid_report_path=self.format_covid_report_path(case=cases[0], ticket_id=ticket_id),
                covid_destination_path=self.format_covid_destination_path(
                    self.covid_destination_path, customer_id=customer_id
                ),
                log_dir=self.log_dir,
            )
        else:
            commands = RSYNC_COMMAND.format(
                source_path=source_and_destination_paths["delivery_source_path"],
                destination_path=source_and_destination_paths["rsync_destination_path"],
            )
        if self.account in SLURM_ACCOUNT_TO_QOS.keys():
            priority = SLURM_ACCOUNT_TO_QOS[self.account]
        else:
            priority = "low"
        sbatch_info = {
            "job_name": "_".join([str(ticket_id), "rsync"]),
            "account": self.account,
            "number_tasks": 1,
            "memory": 1,
            "log_dir": self.log_dir.as_posix(),
            "email": self.mail_user,
            "hours": 24,
            "priority": priority,
            "commands": commands,
            "error": ERROR_RSYNC_FUNCTION.format(),
            "exclude": "--exclude=gpu-compute-0-[0-1],cg-dragen",
        }
        slurm_api = SlurmAPI()
        slurm_api.set_dry_run(dry_run=dry_run)
        sbatch_content: str = slurm_api.generate_sbatch_content(Sbatch.parse_obj(sbatch_info))
        sbatch_path = self.log_dir / "_".join([str(ticket_id), "rsync.sh"])
        sbatch_number: int = slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        return sbatch_number



""" tests for the meta BackupAPI """

import logging
import subprocess

import mock
import pytest
from cg.meta.backup import BackupApi


@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_maximum_flowcells_ondisk_reached(mock_store, mock_pdc):
    """tests maximum_flowcells_ondisk method of the backup api"""
    # GIVEN a flowcell needs to be retrieved from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )

    # WHEN the number of flowcells with status "ondisk" greater than the maximum number allowed
    mock_store.flowcells(status="ondisk").count.return_value = 2000

    # THEN this method should return True
    assert backup_api.maximum_flowcells_ondisk() is True


@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_maximum_flowcells_ondisk_not_reached(mock_store, mock_pdc):
    """tests maximum_flowcells_ondisk method of the backup api"""
    # GIVEN a flowcell needs to be retrieved from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )

    # WHEN the number of flowcells with status "ondisk" less than the maximum number allowed
    mock_store.flowcells(status="ondisk").count.return_value = 1000

    # THEN this method should return False
    assert backup_api.maximum_flowcells_ondisk() is False


@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_maximum_processing_queue_full(mock_store, mock_pdc):
    """tests check_processing method of the backup api"""
    # GIVEN a flowcell needs to be retrieved from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )

    # WHEN there's already a flowcell being retrieved from PDC
    mock_store.flowcells(status="processing").count.return_value = 1

    # THEN this method should return False
    assert backup_api.check_processing(max_processing_flowcells=1) is False


@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_maximum_processing_queue_not_full(mock_store, mock_pdc):
    """tests check_processing method of the backup api"""
    # GIVEN a flowcell needs to be retrieved from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )

    # WHEN there are no flowcells being retrieved from PDC
    mock_store.flowcells(status="processing").count.return_value = 0

    # THEN this method should return True
    assert backup_api.check_processing(max_processing_flowcells=1) is True


@mock.patch("cg.store.models.Flowcell")
@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_pop_flowcell_next_requested(mock_store, mock_pdc, mock_flowcell):
    """tests pop_flowcell method of the backup api"""
    # GIVEN status-db needs to be checked for flowcells to be retrieved from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )

    # WHEN a flowcell is requested to be retrieved from PDC
    mock_store.flowcells(status="requested").first.return_value = mock_flowcell

    popped_flowcell = backup_api.pop_flowcell(dry_run=False)

    # THEN a flowcell is returned, the status is set to "processing", and status-db is updated with
    # the new status
    assert popped_flowcell is not None


@mock.patch("cg.store.models.Flowcell")
@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_pop_flowcell_dry_run(mock_store, mock_pdc, mock_flowcell):
    """tests pop_flowcell method of the backup api"""
    # GIVEN status-db needs to be checked for flowcells to be retrieved from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )

    # WHEN a flowcell is requested to be retrieved from PDC
    # AND it's a  dry run
    popped_flowcell = backup_api.pop_flowcell(dry_run=True)

    # THEN a flowcell is returned, the status is set to "processing", but status-db is NOT updated with
    # the new status
    assert popped_flowcell is not None
    assert not mock_store.commit.called


@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_pop_flowcell_no_flowcell_requested(mock_store, mock_pdc):
    """tests pop_flowcell method of the backup api"""
    # GIVEN status-db needs to be checked for flowcells to be retrieved from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )

    # WHEN there are no flowcells requested to be retrieved from PDC
    mock_store.flowcells(status="requested").first.return_value = None

    popped_flowcell = backup_api.pop_flowcell(dry_run=False)

    # THEN no flowcell is returned
    assert popped_flowcell is None


@mock.patch("cg.meta.backup.BackupApi.check_processing")
@mock.patch("cg.store.models.Flowcell")
@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_fetch_flowcell_processing_queue_full(
    mock_store, mock_pdc, mock_flowcell, mock_check_processing, caplog
):
    """tests the fetch_flowcell method of the backup API"""

    caplog.set_level(logging.INFO)

    # GIVEN we check if a flowcell needs to be retrieved from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )

    # WHEN the processing queue is full
    backup_api.check_processing.return_value = False
    result = backup_api.fetch_flowcell(mock_flowcell, dry_run=False)

    # THEN no flowcell will be fetched and a log message indicates that the processing queue is
    # full
    assert result is None
    assert "processing queue is full" in caplog.text


@mock.patch("cg.meta.backup.BackupApi.maximum_flowcells_ondisk")
@mock.patch("cg.meta.backup.BackupApi.check_processing")
@mock.patch("cg.store.models.Flowcell")
@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_fetch_flowcell_max_flowcells_ondisk(
    mock_store,
    mock_pdc,
    mock_flowcell,
    mock_check_processing,
    mock_maximum_flowcells_ondisk,
    caplog,
):
    """tests the fetch_flowcell method of the backup API"""

    caplog.set_level(logging.INFO)

    # GIVEN we check if a flowcell needs to be retrieved from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )

    # WHEN the processing queue is not full but the number of flowcells on disk is greater than the
    # maximum
    backup_api.check_processing.return_value = True
    backup_api.maximum_flowcells_ondisk.return_value = True

    result = backup_api.fetch_flowcell(mock_flowcell, dry_run=False)

    # THEN no flowcell will be fetched and a log message indicates that maximum number of flowcells
    # has been reached
    assert result is None
    assert "maximum flowcells ondisk reached" in caplog.text


@mock.patch("cg.meta.backup.BackupApi.pop_flowcell")
@mock.patch("cg.meta.backup.BackupApi.maximum_flowcells_ondisk")
@mock.patch("cg.meta.backup.BackupApi.check_processing")
@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_fetch_flowcell_no_flowcells_requested(
    mock_store,
    mock_pdc,
    mock_check_processing,
    mock_maximum_flowcells_ondisk,
    mock_pop_flowcell,
    caplog,
):
    """tests the fetch_flowcell method of the backup API"""

    caplog.set_level(logging.INFO)

    # GIVEN we check if a flowcell needs to be retrieved from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )

    # WHEN no flowcells are requested
    mock_pop_flowcell.return_value = None
    backup_api.check_processing.return_value = True
    backup_api.maximum_flowcells_ondisk.return_value = False

    # AND no flowcell has been specified
    mock_flowcell = None

    result = backup_api.fetch_flowcell(mock_flowcell, dry_run=False)

    # THEN no flowcell will be fetched and a log message indicates that no flowcells have been
    # requested
    assert result is None
    assert "no flowcells requested" in caplog.text


@mock.patch("cg.meta.backup.BackupApi.pop_flowcell")
@mock.patch("cg.meta.backup.BackupApi.maximum_flowcells_ondisk")
@mock.patch("cg.meta.backup.BackupApi.check_processing")
@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_fetch_flowcell_retrieve_next_flowcell(
    mock_store,
    mock_pdc,
    mock_check_processing,
    mock_maximum_flowcells_ondisk,
    mock_pop_flowcell,
    caplog,
):
    """tests the fetch_flowcell method of the backup API"""

    caplog.set_level(logging.INFO)

    # GIVEN we check if a flowcell needs to be retrieved from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )

    # WHEN no flowcells is specified, but a flowcell in status-db has the status "requested"
    mock_flowcell = None
    mock_pop_flowcell.return_value = mock_store.add_flowcell(
        status="requested",
    )
    backup_api.check_processing.return_value = True
    backup_api.maximum_flowcells_ondisk.return_value = False

    result = backup_api.fetch_flowcell(mock_flowcell, dry_run=False)

    # THEN the process to retrieve the flowcell from PDC is started
    assert "retrieving from PDC" in caplog.text

    # AND when done the status of that flowcell is set to "retrieved"
    assert (
        f'Status for flowcell {mock_pop_flowcell.return_value.name} set to "retrieved"'
        in caplog.text
    )
    assert mock_pop_flowcell.return_value.status == "retrieved"

    # AND status-db is updated with the new status
    assert mock_store.commit.called

    # AND the elapsed time of the retrieval process is returned
    assert result > 0


@mock.patch("cg.meta.backup.BackupApi.maximum_flowcells_ondisk")
@mock.patch("cg.meta.backup.BackupApi.check_processing")
@mock.patch("cg.store.models.Flowcell")
@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_fetch_flowcell_retrieve_specified_flowcell(
    mock_store,
    mock_pdc,
    mock_flowcell,
    mock_check_processing,
    mock_maximum_flowcells_ondisk,
    caplog,
):
    """tests the fetch_flowcell method of the backup API"""

    caplog.set_level(logging.INFO)

    # GIVEN we want to retrieve a specific flowcell from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )
    backup_api.check_processing.return_value = True
    backup_api.maximum_flowcells_ondisk.return_value = False

    result = backup_api.fetch_flowcell(mock_flowcell, dry_run=False)

    # THEN the process to retrieve the flowcell from PDC is started
    assert "retrieving from PDC" in caplog.text

    # AND when done the status of that flowcell is set to "retrieved"
    assert f'Status for flowcell {mock_flowcell.name} set to "retrieved"' in caplog.text
    assert mock_flowcell.status == "retrieved"

    # AND status-db is updated with the new status
    assert mock_store.commit.called

    # AND the elapsed time of the retrieval process is returned
    assert result > 0


@mock.patch("cg.meta.backup.BackupApi.maximum_flowcells_ondisk")
@mock.patch("cg.meta.backup.BackupApi.check_processing")
@mock.patch("cg.store.models.Flowcell")
@mock.patch("cg.apps.pdc")
@mock.patch("cg.store")
def test_fetch_flowcell_pdc_retrieval_failed(
    mock_store,
    mock_pdc,
    mock_flowcell,
    mock_check_processing,
    mock_maximum_flowcells_ondisk,
    caplog,
):
    """tests the fetch_flowcell method of the backup API"""

    caplog.set_level(logging.INFO)

    # GIVEN we are going to retrieve a flowcell from PDC
    backup_api = BackupApi(
        mock_store, mock_pdc, max_flowcells_on_disk=1250, root_dir="/path/to/root_dir"
    )
    backup_api.check_processing.return_value = True
    backup_api.maximum_flowcells_ondisk.return_value = False

    # WHEN the retrieval process fails
    mock_pdc.retrieve_flowcell.side_effect = subprocess.CalledProcessError(1, "echo")
    with pytest.raises(subprocess.CalledProcessError):
        backup_api.fetch_flowcell(mock_flowcell, dry_run=False)

    # THEN the failure to retrieve is logged
    assert "retrieval failed" in caplog.text
